import 'package:clean_android_design/clean_android_design.dart';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/material.dart';

class UpdateScaffoldWidget<T extends UpdateableDatabaseItem<T>>
    extends StatefulWidget {
  final IUpdateScaffoldWidgetService<T> databaseCollection;
  final T Function() createNewItem;
  final String? updateableItemId;
  final ILockService? lockService;

  final Widget Function(BuildContext context) builder;
  final PreferredSizeWidget? Function(BuildContext context) appBarBuilder;
  final Widget? Function(BuildContext context)? menuBarBuilder;
  final Widget? Function(BuildContext context)? drawerBuilder;
  final IHelpWidget? helpWidget;
  final bool scrollable;

  final Future<int>? Function()? preparationFuture;
  final Widget Function(BuildContext context)? builderPreparation;
  final PreferredSizeWidget? Function(BuildContext context)?
      appBarBuilderPreparation;
  final Widget? Function(BuildContext context)? menuBarBuilderPreparation;
  final Widget? Function(BuildContext context)? drawerBuilderPreparation;

  final Future<bool> Function(BuildContext context)? onBeforeSave;
  final void Function(BuildContext context, DocumentReference? itemReference)?
      onAfterInsert;
  final void Function(BuildContext context)? onAbort;
  final Future<bool> Function(BuildContext context)? onBeforeDelete;

  final void Function(Validator<T> validator)? initValidator;

  final List<String>? fieldsNotToUpdate;

  const UpdateScaffoldWidget({
    Key? key,
    required this.databaseCollection,
    required this.createNewItem,
    required this.updateableItemId,
    required this.builder,
    required this.appBarBuilder,
    this.menuBarBuilder,
    this.drawerBuilder,
    this.helpWidget,
    this.preparationFuture,
    this.builderPreparation,
    this.appBarBuilderPreparation,
    this.menuBarBuilderPreparation,
    this.drawerBuilderPreparation,
    this.onBeforeSave,
    this.onAfterInsert,
    this.onAbort,
    this.initValidator,
    this.fieldsNotToUpdate,
    this.onBeforeDelete,
    required this.lockService,
    this.scrollable = true,
  }) : super(key: key);

  @override
  State<UpdateScaffoldWidget<T>> createState() =>
      _UpdateScaffoldWidgetState<T>();
}

class _UpdateScaffoldWidgetState<T extends UpdateableDatabaseItem<T>>
    extends State<UpdateScaffoldWidget<T>> {
  late Validator<T> validator;
  late UpdateItemFrame<T> updateItemFrame;

  Stream<T>? documentStream;
  Future<T>? firstItemFuture;
  bool gotStreamError = false;

  @override
  void initState() {
    super.initState();
    validator = Validator<T>();
    widget.initValidator?.call(validator);

    updateItemFrame = UpdateItemFrame<T>();

    getItem();
  }

  void getItem() {
    if (widget.updateableItemId == null) {
      updateItemFrame.setNewItem(widget.createNewItem());
      updateItemFrame.setEditable(true);
      Future.delayed(
          Duration.zero, () => LockNavigatorService.instance.lockForTopPage());
    } else {
      documentStream =
          widget.databaseCollection.documentStream(widget.updateableItemId!);
      firstItemFuture = documentStream!.first;
      documentStream!.handleError((_) {
        setState(() => gotStreamError = true);
      }).listen((item) {
        firstItemFuture = null;
        gotStreamError = false;
        updateItemFrame.setNewItem(item);
        setState(() {});
      });
    }
  }

  @override
  Widget build(BuildContext context) {
    return UpdateController<T>(
      updateItemFrame: updateItemFrame,
      validator: validator,
      databaseCollection: widget.databaseCollection,
      lockService: widget.lockService,
      fieldsNotToUpdate: widget.fieldsNotToUpdate,
      onAbort: widget.onAbort,
      onAfterInsert: widget.onAfterInsert,
      onBeforeSave: widget.onBeforeSave,
      onBeforeDelete: widget.onBeforeDelete,
      child: BasicScaffold(
        preparationFuture: preparationFuture,
        appBarBuilder: widget.appBarBuilder,
        appBarBuilderPreparation: widget.appBarBuilderPreparation,
        drawerBuilder: widget.drawerBuilder,
        drawerBuilderPreparation: widget.drawerBuilderPreparation,
        menuBarBuilder: widget.menuBarBuilder,
        menuBarBuilderPreparation: widget.menuBarBuilderPreparation,
        helpWidget: widget.helpWidget,
        onWillPop: onWillPop,
        scrollable: widget.scrollable,
        builder: builder,
        builderPreparation: builderPreparation,
      ),
    );
  }

  Widget builder(BuildContext context) {
    return StatefulBuilder(builder: (context, setState) {
      UpdateController.of<T>(context).updateNotifier.addListener(() {
        if (mounted) setState(() {});
      });
      return widget.builder.call(context);
    });
  }

  Widget builderPreparation(BuildContext context) {
    if (gotStreamError) {
      return Column(
        mainAxisSize: MainAxisSize.min,
        mainAxisAlignment: MainAxisAlignment.start,
        children: [
          Container(height: 80),
          Icon(
            Icons.report_problem,
            size: 50,
            color: DefaultColors.textLight,
          ),
          DescriptionText(
              "Daten konnten nicht geladen werden. Entweder existiert dieses Dokument nicht mehr oder es besteht zurzeit keine Internetvebindung."),
        ],
      );
    } else {
      return widget.builderPreparation?.call(context) ??
          widget.builder.call(context);
    }
  }

  Future<int>? preparationFuture() {
    if (gotStreamError) return Future.delayed(Duration(days: 10), () => 0);

    var parentFuture = widget.preparationFuture?.call();

    if (parentFuture == null && firstItemFuture == null) return null;

    if (parentFuture != null && firstItemFuture != null)
      return Future.wait<int>([parentFuture, firstItemFuture!.then((_) => 1)])
          .then((_) => 1);

    return parentFuture ?? firstItemFuture!.then((_) => 1);
  }

  Future<bool> onWillPop(BuildContext context) async {
    if (!updateItemFrame.isEditable) return true;

    if (updateItemFrame.editableItem.equals(updateItemFrame.originalItem))
      return true;

    bool? save;
    await SplitRouterDelegate.instance.pushDialog(SaveOrAbortDialog(
      onAbort: () {
        save = false;
      },
      onSave: () {
        save = true;
      },
    ));

    if (save == null) return false;

    if (save == true) {
      var saveWasPossible =
          await UpdateController.of<T>(context).trySave(context);
      return saveWasPossible;
    } else {
      await UpdateController.of<T>(context).abortEditing(context);
      return true;
    }
  }
}
